## @package test
# 
#  @author  Rodrigo Pedroso Mendes
#  @version V1.0
#  @date    24/02/23 01:05:02
#
#  #LICENSE# 
#    
#  Copyright (c) 2023 Rodrigo Pedroso Mendes
#
#  Permission is hereby granted, free of charge, to any  person   obtaining  a 
#  copy of this software and associated  documentation files (the "Software"), 
#  to deal in the Software without restriction, including  without  limitation 
#  the rights to use, copy, modify,  merge,  publish,  distribute, sublicense, 
#  and/or sell copies of the Software, and  to  permit  persons  to  whom  the 
#  Software is furnished to do so, subject to the following conditions:        
#   
#  The above copyright notice and this permission notice shall be included  in 
#  all copies or substantial portions of the Software.                         
#   
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,  EXPRESS OR 
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE  WARRANTIES  OF  MERCHANTABILITY, 
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
#  AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY  CLAIM,  DAMAGES  OR  OTHER 
#  LIABILITY, WHETHER IN AN ACTION OF  CONTRACT, TORT  OR  OTHERWISE,  ARISING 
#  FROM, OUT OF OR IN CONNECTION  WITH  THE  SOFTWARE  OR  THE  USE  OR  OTHER  
#  DEALINGS IN THE SOFTWARE. 
#    
################################################################################
import unittest
import sympy as si
from netlist2ss import netlist2ss


class Test(unittest.TestCase):

    ############################################################################
    # Resistor
    ############################################################################
    def testRES(self):
        netlist = ("R1 GND N1  R1\n"
                   "R2 N2  GND R2\n"
                   "R3 N1  N3  R3\n"
                   "I1 N2  N3  I1\n")
        #Define symbols
        R1  = si.simplify('R1')
        R2  = si.simplify('R2')
        R3  = si.simplify('R3')
        I1  = si.simplify('I1')
        #Reference Matrices
        A_ref = si.Matrix(0, 0, [])
        A_ref = si.simplify(A_ref)

        B_ref = si.Matrix(0, 1, [])
        B_ref = si.simplify(B_ref)

        C_ref = si.Matrix(6, 0, [])
        C_ref = si.simplify(C_ref)

        D_ref = si.Matrix([[-R1], \
                           [-R2], \
                           [-R3], \
                           [-1],  \
                           [-1],  \
                           [-1] ])
        D_ref = si.simplify(D_ref)

        DC_OP_ref = si.Matrix([[-R1*I1], \
                               [-R2*I1], \
                               [-R3*I1], \
                               [-I1],    \
                               [-I1],    \
                               [-I1] ])    
        DC_OP_ref = si.simplify(DC_OP_ref)
        # Run test  
        (A, B, C, D, DC_OP) = netlist2ss( netlist, ['I1'], \
                              ['VdR1','VdR2','VdR3','IdR1','IdR2','IdR3'] )
        #Asserts
        self.assertTrue(A.equals(A_ref))
        self.assertTrue(B.equals(B_ref))
        self.assertTrue(C.equals(C_ref))
        self.assertTrue(D.equals(D_ref))
        self.assertTrue(DC_OP.equals(DC_OP_ref))                          
                                                                 
    ############################################################################
    # VSRC
    ############################################################################
    def testVSRC(self):
        netlist = ("R1 N1 N2  R1\n"
                   "R2 N2 GND R2\n"
                   "V1 N1 GND V1\n"
                   "R3 N3 GND R3\n"
                   "V2 N3 N4  V2\n"
                   "R4 N4 GND R4 \n"
                   "R5 N5 N6  R5\n"
                   "R6 N6 GND R6\n"
                   "V3 GND N5 V3\n")
        #Define symbols
        R1  = si.simplify('R1')
        R2  = si.simplify('R2')
        R3  = si.simplify('R3')
        R4  = si.simplify('R4')
        R5  = si.simplify('R5')
        R6  = si.simplify('R6')
        V1  = si.simplify('V1')
        V2  = si.simplify('V2')
        V3  = si.simplify('V3')
        #Reference Matrices
        A_ref = si.Matrix(0, 0, [])
        A_ref = si.simplify(A_ref)

        B_ref = si.Matrix(0, 3, [])
        B_ref = si.simplify(B_ref)

        C_ref = si.Matrix(6, 0, [])
        C_ref = si.simplify(C_ref)

        D_ref = si.Matrix([[            1,            0,            0], \
                           [            0,            1,            0], \
                           [            0,            0,            1], \
                           [ -1/(R1 + R2),            0,            0], \
                           [            0, -1/(R3 + R4),            0], \
                           [            0,            0, -1/(R5 + R6)]])
        D_ref = si.simplify(D_ref)

        DC_OP_ref = si.Matrix([[ V1 ], \
                               [ V2 ], \
                               [ V3 ], \
                               [ -V1/(R1 + R2) ], \
                               [ -V2/(R3 + R4) ], \
                               [ -V3/(R5 + R6) ]])
        DC_OP_ref = si.simplify(DC_OP_ref)
        # Run test  
        (A, B, C, D, DC_OP) = netlist2ss( netlist, ['V1', 'V2', 'V3'], \
                              ['VdV1','VdV2','VdV3','IdV1','IdV2','IdV3'] )


        #Asserts
        self.assertTrue(A.equals(A_ref))
        self.assertTrue(B.equals(B_ref))
        self.assertTrue(C.equals(C_ref))
        self.assertTrue(D.equals(D_ref))
        self.assertTrue(DC_OP.equals(DC_OP_ref))

    ############################################################################
    # ISRC
    ############################################################################
    def testISRC(self):
        netlist = ("R1 N1 N2 R1\n"
                   "R2 N2 GND R2\n"
                   "I1 N1 GND I1\n"
                   "R3 N3 GND R3\n"
                   "I2 N3 N4 I2\n"
                   "R4 N4 GND R4 \n"
                   "R5 N5 N6 R5\n"
                   "R6 N6 GND R6\n"
                   "I3 GND N5 I3\n")
        #Define symbols
        R1  = si.simplify('R1')
        R2  = si.simplify('R2')
        R3  = si.simplify('R3')
        R4  = si.simplify('R4')
        R5  = si.simplify('R5')
        R6  = si.simplify('R6')
        I1  = si.simplify('I1')
        I2  = si.simplify('I2')
        I3  = si.simplify('I3')
        #Reference Matrices
        A_ref = si.Matrix(0, 0, [])
        A_ref = si.simplify(A_ref)

        B_ref = si.Matrix(0, 3, [])
        B_ref = si.simplify(B_ref)

        C_ref = si.Matrix(6, 0, [])
        C_ref = si.simplify(C_ref)

        D_ref = si.Matrix([[ -R1 -R2,        0,       0], \
                           [        0, -R3 -R4,       0], \
                           [        0,       0, -R5 -R6], \
                           [        1,       0,       0], \
                           [        0,       1,       0], \
                           [        0,       0,       1]])
        D_ref = si.simplify(D_ref)

        DC_OP_ref = si.Matrix([[ -I1*(R1 + R2)], \
                               [ -I2*(R3 + R4)], \
                               [ -I3*(R5 + R6)], \
                               [ I1 ], \
                               [ I2 ], \
                               [ I3 ]])
        DC_OP_ref = si.simplify(DC_OP_ref)
        # Run test  
        (A, B, C, D, DC_OP) = netlist2ss( netlist, ['I1', 'I2', 'I3'], \
                              ['VdI1','VdI2','VdI3','IdI1','IdI2','IdI3'] )


        #Asserts
        self.assertTrue(A.equals(A_ref))
        self.assertTrue(B.equals(B_ref))
        self.assertTrue(C.equals(C_ref))
        self.assertTrue(D.equals(D_ref))
        self.assertTrue(DC_OP.equals(DC_OP_ref))    
    
    ############################################################################
    # Capacitor
    ############################################################################
    def testCAP(self):
        netlist = ("VIN N1  GND VIN\n"
                   "R1  N1  N2  R1\n"
                   "C1  N2  GND C1\n"
                   "R2  N1  N3  R2\n"
                   "C2  GND N3  C2\n"
                   "C3  N1  N4  C3\n"
                   "R3  N4  GND R3\n")
        
        #Define symbols
        R1  = si.simplify('R1')
        R2  = si.simplify('R2')
        R3  = si.simplify('R3')
        C1  = si.simplify('C1')
        C2  = si.simplify('C2')
        C3  = si.simplify('C3')
        VIN = si.simplify('VIN')
        
        #Reference Matrices
        A_ref = si.Matrix([[-1/(R1*C1),          0,          0], \
                           [         0, -1/(R2*C2),          0], \
                           [         0,          0, -1/(R3*C3)]])
        A_ref = si.simplify(A_ref)

        B_ref = si.Matrix([[ 1/(R1*C1)], \
                           [-1/(R2*C2)], \
                           [ 1/(R3*C3)]])
        B_ref = si.simplify(B_ref)
    
        C_ref = si.Matrix([[-1/R1,     0,     0], \
                           [    1,     0,     0], \
                           [    0, -1/R2,     0], \
                           [    0,     1,     0], \
                           [    0,     0, -1/R3], \
                           [    0,     0,     1]])
        C_ref = si.simplify(C_ref)
    
        D_ref = si.Matrix([[ 1/R1], \
                           [    0], \
                           [-1/R2], \
                           [    0], \
                           [ 1/R3], \
                           [    0]])
        D_ref = si.simplify(D_ref)

        DC_OP_ref = si.Matrix([[   0], \
                               [ VIN], \
                               [   0], \
                               [-VIN], \
                               [   0], \
                               [ VIN]])
        DC_OP_ref = si.simplify(DC_OP_ref)
        # Run test  
        (A, B, C, D, DC_OP) = netlist2ss( netlist, ['VIN'], \
                              ['IdC1','VdC1','IdC2','VdC2','IdC3','VdC3'] )

        #Asserts
        self.assertTrue(A.equals(A_ref))
        self.assertTrue(B.equals(B_ref))
        self.assertTrue(C.equals(C_ref))
        self.assertTrue(D.equals(D_ref))
        self.assertTrue(DC_OP.equals(DC_OP_ref))
                    
    ############################################################################
    # Current controlled current source
    ############################################################################
    def testCCCS(self):
        netlist = ("I1 N1  GND I1\n"
                   "I2 N2  GND I2\n"
                   "I3 N1a N2a I1\n"
                   "I4 N2a N1a I2\n"
                   "I5 N1b N2b I1\n"
                   "I6 N2b N1b I2\n"
                   "I7 N1c N2c I1\n"
                   "I8 N2c N1c I2\n"
                   "Ra N2a GND 1\n"
                   "Rb N2b GND 1\n"
                   "Rc N2c GND 1\n"
                   "F1 N3  GND N1   GND F1\n"
                   "F2 N4  GND GND  N2  F2\n"
                   "F3 N5  GND N1a  N2a F3\n"
                   "F4 GND N6  N1b  N2b F4\n"
                   "F5 N7  N8  N1c  N2c F5\n"
                   "R1 N3  GND R1\n"
                   "R2 N4  GND R2\n"
                   "R3 N5  GND R3\n"
                   "R4 N6  GND R4\n"
                   "R5 N7  N8  R5\n"
                   "R6 N8  GND R6")
        
        #Define symbols
        F1  = si.simplify('F1')
        F2  = si.simplify('F2')
        F3  = si.simplify('F3')
        F4  = si.simplify('F4')
        F5  = si.simplify('F5')
        R1  = si.simplify('R1')
        R2  = si.simplify('R2')
        R3  = si.simplify('R3')
        R4  = si.simplify('R4')
        R5  = si.simplify('R5')
        I1  = si.simplify('I1')
        I2  = si.simplify('I2')

        #Reference Matrices
        A_ref = si.Matrix(0, 0, [])
        A_ref = si.simplify(A_ref)

        B_ref = si.Matrix(0, 2, [])
        B_ref = si.simplify(B_ref)

        C_ref = si.Matrix(6, 0, [])
        C_ref = si.simplify(C_ref)

        D_ref = si.Matrix([[  F1*R1,      0], \
                           [      0, -F2*R2], \
                           [  F3*R3, -F3*R3], \
                           [ -F4*R4,  F4*R4], \
                           [  F5*R5, -F5*R5], \
                           [    -F5,     F5]])
        D_ref = si.simplify(D_ref)

        DC_OP_ref = si.Matrix([[  I1*F1*R1 ],       \
                               [ -I2*F2*R2 ],       \
                               [ (I1 - I2)*F3*R3 ], \
                               [ (I2 - I1)*F4*R4 ], \
                               [ (I1 - I2)*F5*R5 ], \
                               [ (I2 - I1)*F5    ]])
        DC_OP_ref = si.simplify(DC_OP_ref)

        # Run test  
        (A, B, C, D, DC_OP) = netlist2ss( netlist, ['I1', 'I2'], \
                              ['VdF1','VdF2','VnN5','VnN6','VdF5','IdF5'] )


        #Asserts
        self.assertTrue(A.equals(A_ref))
        self.assertTrue(B.equals(B_ref))
        self.assertTrue(C.equals(C_ref))
        self.assertTrue(D.equals(D_ref))
        self.assertTrue(DC_OP.equals(DC_OP_ref))

    ############################################################################
    # Inductor
    ############################################################################
    def testIND(self):
        netlist = ("I1  N1  GND IIN\n"
                   "R1  N1  GND R1\n"
                   "L1  N1  GND L1\n"
                   "I2  N2  GND IIN\n"
                   "R2  N2  GND R2\n"
                   "L2  GND N2  L2\n")
        #Define symbols
        R1  = si.simplify('R1')
        R2  = si.simplify('R2')
        R3  = si.simplify('R3')
        L1  = si.simplify('L1')
        L2  = si.simplify('L2')
        L3  = si.simplify('L3')
        IIN = si.simplify('IIN')
        #Reference Matrices
        A_ref = si.Matrix([[-R1/L1,      0], \
                           [     0, -R2/L2] ])
        A_ref = si.simplify(A_ref)

        B_ref = si.Matrix([[-R1/L1], \
                           [ R2/L2] ])
        B_ref = si.simplify(B_ref)

        C_ref = si.Matrix([[-R1,   0], \
                           [  1,   0], \
                           [  0, -R2], \
                           [  0,   1] ])
        C_ref = si.simplify(C_ref)

        D_ref = si.Matrix([[-R1], \
                           [  0], \
                           [ R2], \
                           [  0] ])
        D_ref = si.simplify(D_ref)

        DC_OP_ref = si.Matrix([[   0], \
                               [-IIN], \
                               [   0], \
                               [ IIN] ])
        DC_OP_ref = si.simplify(DC_OP_ref)
        # Run test  
        (A, B, C, D, DC_OP) = netlist2ss( netlist, ['IIN'], \
                              ['VdL1','IdL1','VdL2','IdL2'] )


        #Asserts
        self.assertTrue(A.equals(A_ref))
        self.assertTrue(B.equals(B_ref))
        self.assertTrue(C.equals(C_ref))
        self.assertTrue(D.equals(D_ref))
        self.assertTrue(DC_OP.equals(DC_OP_ref))

    ############################################################################
    # Voltage controlled voltage source
    ############################################################################
    def testVCVS(self):
        netlist = ("V1 N1  GND V1\n"
                   "V2 N2  GND V2\n"
                   "E1 N3  GND N1   GND EA\n"
                   "E2 N4  GND GND  N2  EB\n"
                   "E3 N5  GND N1   N2  EC\n"
                   "E4 GND N6  N1   N2  ED\n"
                   "E5 N7  N8  N1   N2  EE\n"
                   "R1 N3  GND R1\n"
                   "R2 N4  GND R2\n"
                   "R3 N5  GND R3\n"
                   "R4 N6  GND R4\n"
                   "R5 N7  N8  R5\n"
                   "R6 N8  GND R6\n")
        #Define symbols
        E1  = si.simplify('EA')
        E2  = si.simplify('EB')
        E3  = si.simplify('EC')
        E4  = si.simplify('ED')
        E5  = si.simplify('EE')
        R5  = si.simplify('R5')
        V1  = si.simplify('V1')
        V2  = si.simplify('V2')

        #Reference Matrices
        A_ref = si.Matrix(0, 0, [])
        A_ref = si.simplify(A_ref)

        B_ref = si.Matrix(0, 2, [])
        B_ref = si.simplify(B_ref)

        C_ref = si.Matrix(6, 0, [])
        C_ref = si.simplify(C_ref)

        D_ref = si.Matrix([[     E1,     0], \
                           [      0,   -E2], \
                           [     E3,   -E3], \
                           [    -E4,    E4], \
                           [     E5,   -E5], \
                           [ -E5/R5, E5/R5]])
        D_ref = si.simplify(D_ref)

        DC_OP_ref = si.Matrix([[  V1*E1          ], \
                               [ -V2*E2          ], \
                               [ (V1 - V2)*E3    ], \
                               [ (V2 - V1)*E4    ], \
                               [ (V1 - V2)*E5    ], \
                               [ (V2 - V1)*E5/R5 ]])
        DC_OP_ref = si.simplify(DC_OP_ref)
        # Run test  
        (A, B, C, D, DC_OP) = netlist2ss( netlist, ['V1', 'V2'], \
                              ['VdE1','VdE2','VnN5','VnN6','VdE5','IdE5'] )


        #Asserts
        self.assertTrue(A.equals(A_ref))
        self.assertTrue(B.equals(B_ref))
        self.assertTrue(C.equals(C_ref))
        self.assertTrue(D.equals(D_ref))
        self.assertTrue(DC_OP.equals(DC_OP_ref))

    ############################################################################
    # Ideal transformer
    ############################################################################
    def testTRAFO(self):
        netlist = ("V1 N1  GND V1\n"
                   "R1 N1  N2  R1\n"
                   "T1 N2  GND N3 GND Beta\n"
                   "R2 N3  GND R2")
        #Define symbols
        Beta  = si.simplify('Beta')
        R2     = si.simplify('R2')
        R1     = si.simplify('R1')
        V1     = si.simplify('V1')

        #Reference Matrices
        A_ref = si.Matrix(0, 0, [])
        A_ref = si.simplify(A_ref)

        B_ref = si.Matrix(0, 1, [])
        B_ref = si.simplify(B_ref)

        C_ref = si.Matrix(3, 0, [])
        C_ref = si.simplify(C_ref)

        D_ref = si.Matrix([[0], \
                           [(1-R1/(R2/(Beta*Beta) + R1))*Beta], \
                           [1/(R2/(Beta*Beta) + R1)/Beta]])
        D_ref = si.simplify(D_ref)

        DC_OP_ref = si.Matrix([[0], \
                               [(1-R1/(R2/(Beta*Beta) + R1))*Beta*V1], \
                               [V1/(R2/(Beta*Beta) + R1)/Beta]])

        DC_OP_ref = si.simplify(DC_OP_ref)
        # Run test  

        (A, B, C, D, DC_OP) = netlist2ss( netlist, ['V1'], ['IdT1', 'VnN3', 'IdR2'] )


        #Asserts
        self.assertTrue(A.equals(A_ref))
        self.assertTrue(B.equals(B_ref))
        self.assertTrue(C.equals(C_ref))
        self.assertTrue(D.equals(D_ref))
        self.assertTrue(DC_OP.equals(DC_OP_ref))

    ############################################################################
    # Current controlled voltage source
    ############################################################################
    def testCCVS(self):
        netlist = ("I1 N1  GND I1\n"
                   "I2 N2  GND I2\n"
                   "I3 N1a N2a I1\n"
                   "I4 N2a N1a I2\n"
                   "I5 N1b N2b I1\n"
                   "I6 N2b N1b I2\n"
                   "I7 N1c N2c I1\n"
                   "I8 N2c N1c I2\n"
                   "Ra N2a GND 1\n"
                   "Rb N2b GND 1\n"
                   "Rc N2c GND 1\n"
                   "H1 N3  GND N1   GND  H1\n"
                   "H2 N4  GND GND  N2   H2\n"
                   "H3 N5  GND N1a  N2a  H3\n"
                   "H4 GND N6  N1b  N2b  H4\n"
                   "H5 N7  N8  N1c  N2c  H5\n"
                   "R1 N3  GND R1\n"
                   "R2 N4  GND R2\n"
                   "R3 N5  GND R3\n"
                   "R4 N6  GND R4\n"
                   "R5 N7  N8  R5\n"
                   "R6 N8  GND R6\n")
        #Define symbols
        H1  = si.simplify('H1')
        H2  = si.simplify('H2')
        H3  = si.simplify('H3')
        H4  = si.simplify('H4')
        H5  = si.simplify('H5')
        R5  = si.simplify('R5')
        I1  = si.simplify('I1')
        I2  = si.simplify('I2')

        #Reference Matrices
        A_ref = si.Matrix(0, 0, [])
        A_ref = si.simplify(A_ref)

        B_ref = si.Matrix(0, 2, [])
        B_ref = si.simplify(B_ref)

        C_ref = si.Matrix(6, 0, [])
        C_ref = si.simplify(C_ref)

        D_ref = si.Matrix([[    -H1,      0], \
                           [      0,     H2], \
                           [    -H3,     H3], \
                           [     H4,    -H4], \
                           [    -H5,     H5], \
                           [  H5/R5, -H5/R5]])
        D_ref = si.simplify(D_ref)

        DC_OP_ref = si.Matrix([[ -I1*H1 ],       \
                               [  I2*H2 ],       \
                               [ (I2 - I1)*H3    ], \
                               [ (I1 - I2)*H4    ], \
                               [ (I2 - I1)*H5    ], \
                               [ (I1 - I2)*H5/R5 ]])
        DC_OP_ref = si.simplify(DC_OP_ref)
        # Run test  
        (A, B, C, D, DC_OP) = netlist2ss( netlist, ['I1', 'I2'], \
                              ['VdH1','VdH2','VnN5','VnN6','VdH5','IdH5'] )


        #Asserts
        self.assertTrue(A.equals(A_ref))
        self.assertTrue(B.equals(B_ref))
        self.assertTrue(C.equals(C_ref))
        self.assertTrue(D.equals(D_ref))
        self.assertTrue(DC_OP.equals(DC_OP_ref))

    ############################################################################
    # voltage controlled current source
    ############################################################################
    def testVCCS(self):
        netlist = ("V1 N1  GND V1\n"
                   "V2 N2  GND V2\n"
                   "G1 N3  GND N1   GND G1\n"
                   "G2 N4  GND GND  N2  G2\n"
                   "G3 N5  GND N1   N2  G3\n"
                   "G4 GND N6  N1   N2  G4\n"
                   "G5 N7  N8  N1   N2  G5\n"
                   "R1 N3  GND R1\n"
                   "R2 N4  GND R2\n"
                   "R3 N5  GND R3\n"
                   "R4 N6  GND R4\n"
                   "R5 N7  N8  R5\n"
                   "R6 N8  GND R6\n")
        #Define symbols
        G1  = si.simplify('G1')
        G2  = si.simplify('G2')
        G3  = si.simplify('G3')
        G4  = si.simplify('G4')
        G5  = si.simplify('G5')
        R1  = si.simplify('R1')
        R2  = si.simplify('R2')
        R3  = si.simplify('R3')
        R4  = si.simplify('R4')
        R5  = si.simplify('R5')
        V1  = si.simplify('V1')
        V2  = si.simplify('V2')

        #Reference Matrices
        A_ref = si.Matrix(0, 0, [])
        A_ref = si.simplify(A_ref)

        B_ref = si.Matrix(0, 2, [])
        B_ref = si.simplify(B_ref)

        C_ref = si.Matrix(6, 0, [])
        C_ref = si.simplify(C_ref)

        D_ref = si.Matrix([[ -G1*R1,      0], \
                           [      0,  G2*R2], \
                           [ -G3*R3,  G3*R3], \
                           [  G4*R4, -G4*R4], \
                           [ -G5*R5,  G5*R5], \
                           [     G5,    -G5]])
        D_ref = si.simplify(D_ref)

        DC_OP_ref = si.Matrix([[ -V1*G1*R1 ], \
                               [  V2*G2*R2 ], \
                               [ (V2 - V1)*G3*R3 ], \
                               [ (V1 - V2)*G4*R4 ], \
                               [ (V2 - V1)*G5*R5 ], \
                               [ (V1 - V2)*G5    ]])
        DC_OP_ref = si.simplify(DC_OP_ref)
        # Run test  
        (A, B, C, D, DC_OP) = netlist2ss( netlist, ['V1', 'V2'], \
                              ['VdG1','VdG2','VnN5','VnN6','VdG5','IdG5'] )


        #Asserts
        self.assertTrue(A.equals(A_ref))
        self.assertTrue(B.equals(B_ref))
        self.assertTrue(C.equals(C_ref))
        self.assertTrue(D.equals(D_ref))
        self.assertTrue(DC_OP.equals(DC_OP_ref))
        
        
                                                                                                                                                     
if __name__ == '__main__':
    unittest.main()
    




#!/usr/bin/python3
#  Documentation for this module.
# 
#  @author  Rodrigo Pedroso Mendes
#  @version V1.0
#  @date    27/01/19 20:19:11
#
#  #LICENSE# 
#    
#  Copyright (c) 2019 Rodrigo Pedroso Mendes
#
#  Permission is hereby granted, free of charge, to any  person   obtaining  a 
#  copy of this software and associated  documentation files (the "Software"), 
#  to deal in the Software without restriction, including  without  limitation 
#  the rights to use, copy, modify,  merge,  publish,  distribute, sublicense, 
#  and/or sell copies of the Software, and  to  permit  persons  to  whom  the 
#  Software is furnished to do so, subject to the following conditions:        
#   
#  The above copyright notice and this permission notice shall be included  in 
#  all copies or substantial portions of the Software.                         
#   
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,  EXPRESS OR 
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE  WARRANTIES  OF  MERCHANTABILITY, 
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
#  AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY  CLAIM,  DAMAGES  OR  OTHER 
#  LIABILITY, WHETHER IN AN ACTION OF  CONTRACT, TORT  OR  OTHERWISE,  ARISING 
#  FROM, OUT OF OR IN CONNECTION  WITH  THE  SOFTWARE  OR  THE  USE  OR  OTHER  
#  DEALINGS IN THE SOFTWARE. 
#    
#  #DESCRIPTION#
#
#  Calculate the transfer function the circuit represented by a netlist
#
################################################################################


#-------------------------------------------------------------------------------
# Mocules do import
#-------------------------------------------------------------------------------
import sys
from   netlist2ss import netlist2ss
import sympy as  si

#-------------------------------------------------------------------------------
# CLI
#-------------------------------------------------------------------------------
def cli():
    #---------------------------------------------------------------------------
    # Check Arguments 
    #---------------------------------------------------------------------------
    if (len(sys.argv) != 4):
        print ('Usage:')
        print ('    ' + sys.argv[0] + \
               ' filename input_variable output_measurement') 
        exit(-1)
    
    #---------------------------------------------------------------------------
    # Read netlist 
    #---------------------------------------------------------------------------
    try:
        handle  = open(sys.argv[1], 'r')
        netlist = handle.read()
        handle.close()
    except:
        print ("File IO Exception")
        exit(-1)

    #---------------------------------------------------------------------------
    # Run netlist2ss and calculate the transfer function
    #---------------------------------------------------------------------------
    s = si.symbols('s')
    try:
        (A, B, C, D, DC_OP) = netlist2ss(netlist, [sys.argv[2]], [sys.argv[3]])
    except Exception as e:
        print("Error: " + str(e))
        exit(-1)
    H = si.simplify(C*((s*(si.eye(A.shape[0]))-A).inv())*B + D)[0,0]
    n, d = si.fraction(H)   

    #---------------------------------------------------------------------------
    # Print result
    #---------------------------------------------------------------------------
    n = str(n)                        
    d = str(d)
    sizen = len(n)
    sized = len(d)
    sizet = max(sizen, sized) + 2 
    print(" ") 
    print(" Transfer function of " + sys.argv[1])
    print(" Input variable: "  + sys.argv[2]) 
    print(" Output variable: " + sys.argv[3][0:2] + "(" + sys.argv[3][2:] + ")")
    print(" ") 
    print("        " + " "*round((sizet - sizen)/2) + n) 
    print(" H(s) = " + "-"*sizet)
    print("        " + " "*round((sizet - sized)/2) + d) 
    print(" ") 

    exit(0)
    
    
#-------------------------------------------------------------------------------
# Main
#-------------------------------------------------------------------------------
if __name__ == "__main__":
    cli()



## @package vagen
#  VerilogA generator
#

##
#  @mainpage  VAGEN - VerilogA generator
#
#  @section description_main Description
#  vagen is a verilogA  generator that can be used for  transient verification
#  and modeling of complex analog IPs. It provides a   wide range   of   basic 
#  models with voltage sources, current sources,  digital  interfaces, clocks, 
#  switches, and source measure units that you  can  use  to  build your model 
#  upon.
#
#  @section license_main License
#     
#  @author  Rodrigo Pedroso Mendes
#  @version V1.0
#  @date    14/02/23 13:37:31
#    
#  Copyright (c) 2023 Rodrigo Pedroso Mendes
#
#  Permission is hereby granted, free of charge, to any  person   obtaining  a 
#  copy of this software and associated  documentation files (the "Software"), 
#  to deal in the Software without restriction, including  without  limitation 
#  the rights to use, copy, modify,  merge,  publish,  distribute, sublicense, 
#  and/or sell copies of the Software, and  to  permit  persons  to  whom  the 
#  Software is furnished to do so, subject to the following conditions:        
#   
#  The above copyright notice and this permission notice shall be included  in 
#  all copies or substantial portions of the Software.                         
#   
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,  EXPRESS OR 
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE  WARRANTIES  OF  MERCHANTABILITY, 
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
#  AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY  CLAIM,  DAMAGES  OR  OTHER 
#  LIABILITY, WHETHER IN AN ACTION OF  CONTRACT, TORT  OR  OTHERWISE,  ARISING 
#  FROM, OUT OF OR IN CONNECTION  WITH  THE  SOFTWARE  OR  THE  USE  OR  OTHER  
#  DEALINGS IN THE SOFTWARE. 
#    
################################################################################

#-------------------------------------------------------------------------------
# Imports
#-------------------------------------------------------------------------------
from netlist2ss.netlist2ss import netlist2ss 





## @package netlist2ss
# 
#  @author  Rodrigo Pedroso Mendes
#  @version V1.0
#  @date    28/01/19 11:41:38
#
#  #LICENSE# 
#    
#  Copyright (c) 2019 Rodrigo Pedroso Mendes
#
#  Permission is hereby granted, free of charge, to any  person   obtaining  a 
#  copy of this software and associated  documentation files (the "Software"), 
#  to deal in the Software without restriction, including  without  limitation 
#  the rights to use, copy, modify,  merge,  publish,  distribute, sublicense, 
#  and/or sell copies of the Software, and  to  permit  persons  to  whom  the 
#  Software is furnished to do so, subject to the following conditions:        
#   
#  The above copyright notice and this permission notice shall be included  in 
#  all copies or substantial portions of the Software.                         
#   
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,  EXPRESS OR 
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE  WARRANTIES  OF  MERCHANTABILITY, 
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
#  AUTHORS OR COPYRIGHT HOLDERS BE  LIABLE FOR ANY  CLAIM,  DAMAGES  OR  OTHER 
#  LIABILITY, WHETHER IN AN ACTION OF  CONTRACT, TORT  OR  OTHERWISE,  ARISING 
#  FROM, OUT OF OR IN CONNECTION  WITH  THE  SOFTWARE  OR  THE  USE  OR  OTHER  
#  DEALINGS IN THE SOFTWARE. 
#    
#  #DESCRIPTION#
#
#      This module  contains functions to convert a spice netlist into a space 
#  state representation of the system
#
################################################################################

#-------------------------------------------------------------------------------
# import necessary modules
#-------------------------------------------------------------------------------
import re
import numpy as np
import sympy as si

#-------------------------------------------------------------------------------
# Error Class
# An error will be raised whenever an internal error occurs
#-------------------------------------------------------------------------------
class Error(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

#-------------------------------------------------------------------------------
# Component Class
# This class contains basic definition of a spice component
#
# The parameters of the constructor are listed bellow:
# name:  Component name (Example: V1, R1, R2, etc)
# nodes: A list with the name of the nets in  which the component is connected 
#        (Example: ['n1', 'n2', ...])
# value: An expression retrieved  from the 'value' field of  the spice netlist
#        (Example: '1e-2*a+b'). This expression is  parsed and assumed  to  be 
#        the value of the component 
#-------------------------------------------------------------------------------
class component:

    #---------------------------------------------------------------------------
    # Constructor
    #---------------------------------------------------------------------------
    def __init__(self, name, nodes, value): 
        self.type   = name[0].upper()
        self.nodes  = nodes
        try:
            self.value = si.sympify(value)
        except:
            raise Error('Unable to parse "' + value + '"')
        self.e1Idx = None
        self.e2Idx = None
        if (self.type in 'LC'):
            self.st = si.symbols('state_var_' + name)     
        else:
            self.st = None

    #---------------------------------------------------------------------------
    # Get fixed parameters of the component
    #---------------------------------------------------------------------------
    def getType(self):
        return self.type

    def getNodes(self):
        return self.nodes

    def getValue(self):
        return self.value

    def getST(self):
        return self.st

    #---------------------------------------------------------------------------
    # Get and set the indexes in the J matrix   
    #---------------------------------------------------------------------------
    def setE1Idx(self, e1Idx):
        self.e1Idx = e1Idx

    def setE2Idx(self, e2Idx):
        self.e2Idx = e2Idx

    def getE1Idx(self):
        return self.e1Idx

    def getE2Idx(self):
        return self.e2Idx

    #---------------------------------------------------------------------------
    # String callbacks in order facilitate debugging 
    #---------------------------------------------------------------------------
    def __str__(self):
        return "< " + self.type       + ", " + str(self.nodes) + ", " + \
                      str(self.value) + ", " + str(self.e1Idx) + ", " + \
                      str(self.e2Idx) + ", " + str(self.st)    + " >"

    def __repr__(self):
        return "< " + self.type       + ", " + str(self.nodes) + ", " + \
                      str(self.value) + ", " + str(self.e1Idx) + ", " + \
                      str(self.e2Idx) + ", " + str(self.st)    + " >"

#-------------------------------------------------------------------------------
# netlistParser
# Recieves a netlist as a raw string and returns a dictionary where  the  keys 
# are the names of the components and the items are composed  of  instances of  
# the component class.
#
# -Inputs
# netlist:  A spice netlist given as a raw string
# -Outputs
# compDict: The component dictionary 
#-------------------------------------------------------------------------------
def netlistParser(netlist):
    #Regex patterns
    patternEmptyLines = r"^[ \t]*$|^[ \t]*\*.*"
    pattern2tDevices  = r"^[ \t]*([VvIiLlRrCc][A-Za-z0-9_]*)[ \t]+" + \
                        r"([A-Za-z0-9_]+)[ \t]+" + \
                        r"([A-Za-z0-9_]+)[ \t]+" + \
                        r"([A-Za-z0-9+*/() -]+)[ \t]*(;.*)?$"
    pattern4tDevices  = r"^[ \t]*([EeHhGgFfTt][A-Za-z0-9_]*)[ \t]+"   + \
                        r"([A-Za-z0-9_]+)[ \t]+" + \
                        r"([A-Za-z0-9_]+)[ \t]+" + \
                        r"([A-Za-z0-9_]+)[ \t]+" + \
                        r"([A-Za-z0-9_]+)[ \t]+" + \
                        r"([A-Za-z0-9+*/() -]+)[ \t]*(;.*)?$"
    #Split the netlist in lines and fillout the component dictionary
    compDict = {}
    compList = []
    lineList = re.findall("(.*)\n*", netlist)
    for line in lineList:
        #Remove empty line and comment
        if(re.search(patternEmptyLines, line)):
            continue
        #Find 2 terminal devices
        elif(re.search(pattern2tDevices, line)):
            compDesc = re.findall(pattern2tDevices, line)[0]
            if compDesc[0] in compDict.keys():
                raise Error(compDesc[0] + ": Duplicated device.")
            compDict[compDesc[0]] = component(compDesc[0],               \
                                              [compDesc[1],compDesc[2]], \
                                              compDesc[3])
            compList.append(compDict[compDesc[0]])
        #Find 4 terminal devices
        elif(re.search(pattern4tDevices, line)):
            compDesc = re.findall(pattern4tDevices, line)[0]
            if compDesc[0] in compDict.keys():
                raise Error(compDesc[0] + ": Duplicated device.")
            compDict[compDesc[0]] = component(compDesc[0],      \
                                    [compDesc[1], compDesc[2],  \
                                     compDesc[3],compDesc[4]],  \
                                    compDesc[5])
            compList.append(compDict[compDesc[0]])
        #Invalid line
        else:
            raise Error("Error when processing the line \"" + line +
                        "\". Unsuported device, net name or device  value")

    return (compDict, compList)

#-------------------------------------------------------------------------------
# calcNodesnJ
# Calculate the size of the J matrix in the nodal analysis, update the indexes
# that each component ocupies in the J matrix,  and transforms the  net  names 
# in a sequence of integer that represents the node number.
#
# -Inputs
# compDict:  The component dictionary generated by the netlistParser
# -Outputs
# nJ:        Size of the J matrix
# nNodes:    Number of nodes in the nodal analysis
# nodesDict: Dictionary corelating the net name with the node number
#
# This function also updates the J matrix index of each afected  component  in
# the component list  
#-------------------------------------------------------------------------------
def calcNodesnJ(compList):
    nJ         = 0
    nNodes     = 0
    nodesDict  = {}
    #Loop trough all components 
    for comp in compList:
        #Independent  voltage  sources, voltage   controled  voltage  sources, 
        #current controled current sources, and capacitors
        if comp.getType() in 'VFEC':
            comp.setE1Idx(nJ)
            nJ = nJ + 1
        #Current controled voltage sources 
        elif comp.getType() in 'HT':
            comp.setE1Idx(nJ)
            nJ = nJ + 1
            comp.setE2Idx(nJ)
            nJ = nJ + 1
        #Loop trough all nodes
        for node in comp.getNodes():
            if not (node in nodesDict.keys()):
                if (node.upper() == 'GND' or node == '0'):
                    nodesDict[node] = -1
                else:
                    nodesDict[node] = nNodes   
                    nNodes = nNodes + 1
    return (nJ, nNodes, nodesDict) 

#-------------------------------------------------------------------------------
# nodalAnalysisMatrices
# Construct the nodal analysis matrices
#
# -Inputs
# compDict:  The component dictionary generated by the netlistParser
# nJ:        Size of the J matrix
# nNodes:    Number of nodes in the nodal analysis
# nodesDict: Dictionary corelating the net name with the node number 
# -Outputs:
# A: The A matrix is a concatenation of the G, B, C, and D matrices
# Z: Z is a column vector which is the concatenation of I and E matrices
#-------------------------------------------------------------------------------
def nodalAnalysisMatrices(compList, nJ, nNodes, nodesDict):

    #---------------------------------------------------------------------------
    # Alocate matrices
    #---------------------------------------------------------------------------
    G = si.zeros(nNodes, nNodes)
    I = si.zeros(nNodes, 1)
    if nJ != 0:
        B = si.zeros(nNodes, nJ)
        C = si.zeros(nJ, nNodes)
        D = si.zeros(nJ, nJ)
        E = si.zeros(nJ, 1)

    #---------------------------------------------------------------------------
    # Loop trough all components 
    #---------------------------------------------------------------------------
    for comp in compList:

        #-----------------------------------------------------------------------
        # Read nodes and J matrix indexes
        #-----------------------------------------------------------------------
        nodes = comp.getNodes()
        n1 = nodesDict[nodes[0]]
        n2 = nodesDict[nodes[1]] 
        if len(nodes) > 2:
            n3 = nodesDict[nodes[2]]
            n4 = nodesDict[nodes[3]] 
        e1 = comp.getE1Idx()
        e2 = comp.getE2Idx()

        #-----------------------------------------------------------------------
        # fill the G matrix with resistors from the netlist
        #-----------------------------------------------------------------------
        if comp.getType() == 'R':
            #If  neither  side  of  the element is  connected  to  ground then 
            #subtract it from appropriate location in matrix.
            if n1 != -1 and n2 != -1:
                G[n1, n2] = G[n1, n2] - 1/comp.getValue()
                G[n2, n1] = G[n2, n1] - 1/comp.getValue()
            #If node 1 is not connected to ground, add element to  diagonal of 
            #matrix
            if n1 != -1:
                G[n1, n1] = G[n1, n1] + 1/comp.getValue()
            #Ditto for node 2.
            if n2 != -1:
                G[n2, n2] = G[n2, n2] + 1/comp.getValue()

        #-----------------------------------------------------------------------
        # fill the G  matrix  with  voltage  controled  current  sources  from 
        # netlist
        #-----------------------------------------------------------------------
        elif comp.getType() == 'G':
            #If neither n1 and n3 is  connected  to  ground  then  sum  it  to 
            #appropriate location in matrix.
            if n1 != -1 and n3 != -1:
                G[n1, n3] = G[n1, n3] + comp.getValue()
            #If neither n1 and n4 is connected to ground then subtract it from
            #appropriate location in matrix.
            if n1 != -1 and n4 != -1:
                G[n1, n4] = G[n1, n4] - comp.getValue()
            #If neither n2 and n3 is connected to ground then subtract it from
            #appropriate location in matrix.
            if n2 != -1 and n3 != -1:
                G[n2, n3] = G[n2, n3] - comp.getValue()
            #If neither n2 and  n4  is  connected to ground  then  sum  it  to 
            #appropriate location in matrix.
            if n2 != -1 and n4 != -1:
                G[n2, n4] = G[n2, n4] + comp.getValue()

        #-----------------------------------------------------------------------
        # fill the I matrix with independent  current  sources  and  inductors 
        # from the netlist.
        #-----------------------------------------------------------------------
        elif comp.getType() in 'I':
            #Check if node 1 is connected to ground, then subtract the current 
            #in the correct location in the matrix 
            if n1 != -1:
                I[n1] = I[n1] - comp.getValue()
            #Check if node 2 is  connected to ground, then sum the  current in
            #the correct location in the matrix 
            if n2 != -1:
                I[n2] = I[n2] + comp.getValue()
        
        #-----------------------------------------------------------------------
        # fill the I matrix with inductors from the netlist.
        # In order to calculate the space  state matrices, we solve  the nodal 
        # analysis system as a function of the inductor current (which is  one
        # of the states), and then calculate  the voltage across the inductor, 
        # which will give us the derivative of the current.
        #-----------------------------------------------------------------------
        elif comp.getType() in 'L':
            #Check if node 1 is connected to ground, then subtract the current
            #in the correct location in the matrix 
            if n1 != -1:
                I[n1] = I[n1] - comp.getST()
            #Check if node 2  is  connected to ground, then sum the current in 
            #the correct location in the matrix 
            if n2 != -1:
                I[n2] = I[n2] + comp.getST()

        #-----------------------------------------------------------------------
        # Fill the C  and  B  matrices  with  independent voltage  sources and 
        # capacitors from the netlist.
        #-----------------------------------------------------------------------
        elif comp.getType() in 'V':
            #Check if node 1 is connected to ground, then fill-out the B and C 
            #matrix
            if n1 != -1:  
                B[n1, e1] = 1      
                C[e1, n1] = 1 
            #Check if node 2 is connected to ground, then fill-out the B and C
            #matrix
            if n2 != -1:  
                B[n2, e1] = -1     
                C[e1, n2] = -1
            #E matrix
            E[e1] = comp.getValue()

        #-----------------------------------------------------------------------
        # Fill the C and B matrices with capacitors from the netlist.
        # In order to calculate the space state  matrices, we solve the  nodal
        # analysis system as a function of the capacitor voltage (which is one 
        # of the states), and then calculate the current through the capacitor 
        # which will give us the derivative of the voltage
        #-----------------------------------------------------------------------
        elif comp.getType() in 'C':
            #Check if node 1 is connected to ground, then fill-out the B and C 
            #matrix
            if n1 != -1:  
                B[n1, e1] = 1      
                C[e1, n1] = 1 
            #Check if node 2 is connected to ground, then fill-out the B and C
            #matrix
            if n2 != -1:  
                B[n2, e1] = -1     
                C[e1, n2] = -1
            #E matrix
            E[e1] = comp.getST()

        #-----------------------------------------------------------------------
        # Fill the B and C matrices with  voltage controlled  voltage  sources
        #-----------------------------------------------------------------------
        elif comp.getType() in 'E':
            #Check if node 1 is connected to ground, then fill-out the B and C 
            #matrix
            if n1 != -1:  
                B[n1, e1] = 1      
                C[e1, n1] = 1 
            #Check if node 2 is connected to ground, then fill-out the B and C
            #matrix
            if n2 != -1:  
                B[n2, e1] = -1     
                C[e1, n2] = -1
            #Check if node 3  is  connected  to  ground, then  fill-out  the C 
            #matrix
            if n3 != -1:      
                C[e1, n3] = -comp.getValue()
            #Check if node 4  is  connected  to  ground, then fill-out  the  C 
            #matrix
            if n4 != -1:      
                C[e1, n4] = comp.getValue()

        #-----------------------------------------------------------------------
        # Fill the B and C matrices  with current controlled  current  sources
        #-----------------------------------------------------------------------
        elif comp.getType() in 'F':
            #Check if node 1  is  connected  to  ground, then fill-out  the  B
            #matrix
            if n1 != -1:  
                B[n1, e1] = comp.getValue() 
            #Check if node 2  is  connected  to  ground, then fill-out  the  B
            #matrix
            if n2 != -1:  
                B[n2, e1] = -comp.getValue()
            #Check if ctrl 1  is  connected  to  ground, then fill-out  the  B
            #and C matrix
            if n3 != -1:   
                B[n3, e1] = 1   
                C[e1, n3] = 1
            #Check if ctrl 2  is  connected  to  ground, then fill-out  the  B
            #and C matrix
            if n4 != -1:   
                B[n4, e1] = -1   
                C[e1, n4] = -1

        #-----------------------------------------------------------------------
        # Fill the B and C matrices with current  controlled  voltage  sources 
        #-----------------------------------------------------------------------
        elif comp.getType() in 'H':
            #Check if node 1 is connected to ground, then fill-out the B and C
            #matrix
            if n1 != -1:  
                B[n1, e2] = 1
                C[e1, n1] = 1
            #Check if node 2 is connected to ground, then fill-out the B and C
            #matrix
            if n2 != -1:  
                B[n2, e2] = -1
                C[e1, n2] = -1
            #Check if node 3 is connected to ground, then fill-out the B and C
            #matrix
            if n3 != -1:   
                B[n3, e1] = 1   
                C[e2, n3] = 1
            #Check if node 4 is connected to ground, then fill-out the B and C
            #matrix
            if n4 != -1:   
                B[n4, e1] = -1   
                C[e2, n4] = -1
            #D matrix
            D[e1, e1] = -comp.getValue()

        #-----------------------------------------------------------------------
        # Ideal transformer 
        #-----------------------------------------------------------------------
        elif comp.getType() in 'T':
            #Check if node 1 is connected to ground, then fill-out the B and C
            #matrix
            if n1 != -1:  
                B[n1, e1] = 1
                C[e2, n1] = -comp.getValue()
            #Check if node 2 is connected to ground, then fill-out the B and C
            #matrix
            if n2 != -1:  
                B[n2, e1] = -1
                C[e2, n2] = comp.getValue()
            #Check if node 3 is connected to ground, then fill-out the B and C
            #matrix
            if n3 != -1:   
                B[n3, e2] = 1   
                C[e2, n3] = 1
            #Check if node 4 is connected to ground, then fill-out the B and C
            #matrix
            if n4 != -1:   
                B[n4, e2] = -1   
                C[e2, n4] = -1
            #D matrix
            D[e1, e2] = comp.getValue()
            D[e1, e1] = 1

    #---------------------------------------------------------------------------
    # Calculate the A and Z matrices 
    #---------------------------------------------------------------------------
    if nJ != 0:
        A = (G.row_join(B)).col_join(C.row_join(D))
        Z = I.col_join(E)
    else:
        A = G
        Z = I

    return (A, Z)  

#-------------------------------------------------------------------------------
# solveSystem
# Solve the symbolic nodal analysis system 
#
# -Inputs
# A:      The A matrix is a concatenation of the G, B, C, and D matrices
# Z:      Z is a column vector which is the concatenation of I and E  matrices
# nNodes: Number of nodes in the nodal analysis
# -Outputs
# V:      The V matrix contains the voltage in all nodes of the system
# J:      The J matrix contains the current flowing trough  current  controled 
#         current  sources,  current   controled   voltage   sources,  voltage 
#         controled  voltage   sources,   independent   voltage   source,  and 
#         capacitors
#-------------------------------------------------------------------------------
def solveSystem(A, Z, nNodes):
    try:
        X = A.inv()*Z
    except: 
        raise Error('Unable to solve the linear system. Check the netlist')
    V = X[0:nNodes, 0]
    J = X[nNodes: , 0]
    V = V.col_join(si.zeros(1, 1))
    return (V, J)

#-------------------------------------------------------------------------------
# stateEquations
# Returns a vector listing all the states and a vector containing  the  set of 
# equations for each state
#
# -Inputs
# compDict:  The component dictionary generated by the netlistParser
# nodesDict: Dictionary corelating the net name with the node number 
# V: The V matrix contains the voltage in all nodes of the system
# J: The J matrix  contains  the  current  flowing  trough  current  controled 
#    current sources, current controled  voltage  sources,  voltage  controled  
#    voltage sources, independent voltage source, and capacitors
# -Outputs
# X: column vector listing all states
# F: column vector with a equation for each state
#-------------------------------------------------------------------------------
def stateEquations (compList, nodesDict, V, J):

    #---------------------------------------------------------------------------
    # Initialize variables 
    #---------------------------------------------------------------------------
    X   = []
    F   = []
    nST = 0

    #---------------------------------------------------------------------------
    # Loop trough all components 
    #---------------------------------------------------------------------------
    for comp in compList:
        #Inductor corresponds to one state
        if comp.getType() == 'L': 
            nodes = comp.getNodes()   
            F.append((V[nodesDict[nodes[0]]] - V[nodesDict[nodes[1]]])/ \
                     comp.getValue())
            X.append(comp.getST())
            nST = nST + 1
        #Capacitor corresponds to another state
        elif comp.getType() == 'C':
            F.append(J[comp.getE1Idx()]/comp.getValue())
            X.append(comp.getST())
            nST = nST + 1

    #---------------------------------------------------------------------------
    # reshape the output
    #---------------------------------------------------------------------------
    F = si.Matrix(nST, 1, F)
    X = si.Matrix(nST, 1, X)

    return (X, F)

#-------------------------------------------------------------------------------
# parseOutputs
# Build output equations
#
# -Inputs
# compDict:  The component dictionary generated by the netlistParser
# nodesDict: Dictionary corelating the net name with the node number 
# V: The V matrix contains the voltage in all nodes of the system
# J: The J matrix  contains  the  current  flowing  trough  current  controled 
#    current sources, current controled  voltage  sources,  voltage  controled  
#    voltage sources, independent voltage source, and capacitors
# outputs: A list containing the desired measurements from  which  the  output
#          equations will be built 
# -Outputs
# G : A column vector containing the set of output equations
#-------------------------------------------------------------------------------
def parseOutputs (compDict, nodesDict, V, J, outputs):

    #---------------------------------------------------------------------------
    # Alocate space 
    #---------------------------------------------------------------------------
    G = si.zeros(len(outputs), 1)

    #---------------------------------------------------------------------------
    # Loop trough all outputs      
    #---------------------------------------------------------------------------
    for i in range(0, len(outputs)):
        meas = outputs[i][0:2]
        name = outputs[i][2:]

        #-----------------------------------------------------------------------
        # Node voltage measurement
        #-----------------------------------------------------------------------
        if meas == 'Vn':
            if not name in nodesDict.keys():
                raise Error("Node " + name + " doesn't exist")
            G[i] = V[nodesDict[name]]

        #-----------------------------------------------------------------------
        # Device related measurements
        #-----------------------------------------------------------------------
        else:
            #Check if the device exists in the device list
            if not name in compDict.keys():
                raise Error("Device " + name + " doesn't exist")
            #Get device type and nodes
            compType = compDict[name].getType() 
            if compType == 'T':
                print("Warning: Measurements for transformer weren't implemented. Skipping")
                continue
            nodes    = compDict[name].getNodes()
            n1 = nodesDict[nodes[0]]
            n2 = nodesDict[nodes[1]] 
            if len(nodes) > 2:
                n3 = nodesDict[nodes[2]]
                n4 = nodesDict[nodes[3]]
            #Get the J matrix indexes
            e1 = compDict[name].getE1Idx() 
            e2 = compDict[name].getE2Idx() 
            #Get component value 
            value = compDict[name].getValue()
            #Get state name
            st = compDict[name].getST()
 
            #-------------------------------------------------------------------
            # Measurement of the voltage across a device 
            #-------------------------------------------------------------------
            if meas == 'Vd':  
                if   compType == 'C': 
                    G[i] = st 
                elif compType in 'HFEGIRL': 
                    G[i] = V[n1] - V[n2]     
                elif compType == 'V':
                    G[i] = value
 
            #-------------------------------------------------------------------
            # Measurement of the current flowing in a device
            #-------------------------------------------------------------------
            elif meas == 'Id': 
                if   compType == 'L': 
                    G[i] = st 
                elif compType in 'CVE': 
                    G[i] = J[e1]   
                elif compType == 'H': 
                    G[i] = J[e2]  
                elif compType == 'F': 
                    G[i] = J[e1]*value
                elif compType == 'I':
                    G[i] = value 
                elif compType == 'R': 
                    G[i] = (V[n1] - V[n2])/value
                elif compType == 'G': 
                    G[i] = (V[n3] - V[n4])*value

            #-------------------------------------------------------------------
            # Measurement  of  the  voltage   across  the  control  nodes of a 
            # dependent source
            #-------------------------------------------------------------------        
            elif meas == 'Vc': 
                if not compType in 'EFGH': 
                    raise Error("There isn't control pins in " + name)
                if compType in 'FH': 
                    G[i] = 0
                elif compType in 'EG': 
                    G[i] = V[n3] - V[n4]

            #-------------------------------------------------------------------
            # Measurement  of  the  current  across  the  control   nodes of a 
            # dependent source
            #-------------------------------------------------------------------  
            elif meas == 'Ic': 
                if not compType in 'EFGH': 
                    raise Error("There isn't control pins in " + name)
                if compType in 'EG': 
                    G[i] = 0
                elif compType == 'F': 
                    G[i] = J[e1]  
                elif compType == 'H':
                    G[i] = J[e1] 

            #-------------------------------------------------------------------
            # Unknown measurement tyoe
            #-------------------------------------------------------------------  
            else:
                raise Error("Unknown measuring type: " + meas)
    return G

#-------------------------------------------------------------------------------
# parseInputs
# build input equations
# 
# -Inputs 
# inputs: a list containing the name of variables consired to be the inputs of
#         the system
# -Output
# U: column vector listing all inputs of the system
#-------------------------------------------------------------------------------
def parseInputs (inputs):
    U = si.Matrix(len(inputs), 1, si.sympify(inputs)) 
    return U

#-------------------------------------------------------------------------------
# calcABCD
# Calculate A, B, C, D, and DC_OP matrices
#
# -Inputs
# F: column vector with a equation for each state
# X: column vector listing all states
# G: column vector containing the set of output equations 
# U: column vector listing all inputs of the system
# -Outputs
# A: state matrix
# B: input matrix
# C: output matrix
# D: feedforward matrix
# DC_OP: operating point
#-------------------------------------------------------------------------------
def calcABCD (F, X, G, U):
    #Calculate the state at the operating point
    nST  = len(X)
    sol  = si.solve(F, X)
    #dummy fix for the empty solution case
    if sol == [] and nST != 0:
        raise Error("The isn't a single solution. Check the netlist.")
    for stVar in X:
        if not stVar in sol.keys():
            raise Error("The isn't a single solution. Check the netlist.")
    X_OP = si.Matrix(nST, 1, [sol[X[i]] for i in range(0, nST)])
    #Linearized system (if the system is not linear) at the operating point 
    A = F.jacobian(X).subs([ (X[i], X_OP[i]) for i in range(0, nST) ])
    A = si.simplify(A)
    B = F.jacobian(U).subs([ (X[i], X_OP[i]) for i in range(0, nST) ])
    B = si.simplify(B)
    C = G.jacobian(X).subs([ (X[i], X_OP[i]) for i in range(0, nST) ])
    C = si.simplify(C)
    D = G.jacobian(U).subs([ (X[i], X_OP[i]) for i in range(0, nST) ])
    D = si.simplify(D)
    #Calculate the outputs of the system at the operating point
    DC_OP = G.subs([ (X[i], X_OP[i]) for i in range(0, nST) ])
    DC_OP = si.simplify(DC_OP)
    return (A, B, C, D, DC_OP)
       
#-------------------------------------------------------------------------------
# netlist2ss
# Convert a netlist to a space state representation of the system.
#  
# -Inputs
# netlist:  A string with a spice netlist
# inputs:   a list containing the name of variables consired to be the  inputs 
#           of the system
# outputs:  A list containing the desired measurements from  which  the output
#           equations will be built 
# -Outputs
# A: state matrix
# B: input matrix
# C: output matrix
# D: feedforward matrix
# DC_OP: operating point   
#
# -example: 
# (A, B, C, D, OP) = netlist2ss("e1 n1 gnd in\nr1 n1 c1 r1\n 
#                                c1 c1 gnd c1", ["in"], ["Vnc1"])
#-------------------------------------------------------------------------------
def netlist2ss(netlist, inputs, outputs):
    #Parse netlist
    (compDict, compList) = netlistParser(netlist)
    #Calculate the number of nodes and the size of the J matrix
    (nJ, nNodes, nodesDict) = calcNodesnJ(compList)
    #Build the nodal analysis matrices
    (A, Z) = nodalAnalysisMatrices(compList, nJ, nNodes, nodesDict)
    #Solve the linear system
    (V, J) = solveSystem(A, Z, nNodes)
    #Select which one of the nodal analysis results are state equations
    (X, F) = stateEquations (compList, nodesDict, V, J)
    #Select which one of the nodal analysis results are output equations 
    G = parseOutputs (compDict, nodesDict, V, J, outputs)
    #Input variables
    U = parseInputs (inputs)
    #Return the space state representation of the system
    return calcABCD (F, X, G, U)
